/**************************
 * Includes
 *
 **************************/

#include <windows.h>
#include <gl/gl.h>
#include <iostream>
#include <math.h>
#include <fstream>


#define PI 3.14159265

using namespace std;

/**************************
 * Function Declarations
 *
 **************************/

void ProjectionTo(float *buff, float x, float y, float dir)
{
    float dist = sqrt(x*x + y*y);
    float angle = dir*PI/180.0 - atan(x/y);
    buff[0] = sin(angle) * dist;
    buff[1] = cos(angle) * dist;
}
class Map;

class Car
{
    float dir;
    float speed;
    float lth_half;
    float wth_half;
    float x;
    float y;
    float angle;
    float dist;
public:
    Car(float nx, float ny,float nwth, float nlth)
    {
        x = nx;
        y = ny;
        dir = 0.0;
        speed = 0.0;
        lth_half = nlth;
        wth_half = nwth;
        dist = sqrt(lth_half*lth_half +wth_half*wth_half);
        angle = atan(wth_half/lth_half);
    }
    void MoveCar(Map *map);


    void DrawCar();


    float getX();


    float getY();

    float getDIR();

    float getDIST();

    float getANGLE();

    float getWTH();

    float getLTH();

};

class House
{
    float x;
    float y;
    float lth_half;
    float wth_half;
    float points[4][2];
public:
    House(float nx, float ny, float nwth, float nlth)
    {
        x = nx;
        y = ny;
        lth_half = nlth;
        wth_half = nwth;
        for(int i = 0; i < 4; i++)
        {
            if(i<2)
                points[i][0] = x - wth_half;
            else
                points[i][0] = x + wth_half;
            if(i%2)
                points[i][1] = y + lth_half;
            else
                points[i][1] = y - lth_half;
        }
    }

    int CollissionDetection(Car *ObjCar);


    void DrawHouse(Car *ObjCar);

};



class MapPiece
{
    House *blds[10];
    int i;
public:
    MapPiece()
    {
        i = 0;
    }

    void push(House *NewHouse);

    int CollisionCheck(Car *ObjCar);

    void DrawPiece(Car *ObjCar);

    ~MapPiece()
    {
        for(int j = 0; j < i; j++)
            delete blds[j];
    }
};



class Map
{
    int mx;
    int my;
    MapPiece **MapArray;
public:
    Map(char *name, Car *ObjCar)
    {
        ifstream map(name);
        map >> mx;
        map >> my;
        MapArray = new MapPiece*[mx];
        for(int i = 0; i < mx; i++)
            MapArray[i] = new MapPiece[my];
        float x, y, lth, wth;
        map >> x;
        map >> y;
        map >> wth;
        map >> lth;
        ObjCar = new Car(x, y, wth, lth);
        while(!map.eof())
        {
            map >> x;
            map >> y;
            map >> wth;
            map >> lth;
            MapArray[(int)(x/1.0)][(int)(y/1.0)].push(new House(x, y, wth, lth));
        }
        map.close();
    }

    int CollisionCheck(Car *ObjCar);

    void DrawMap(Car *ObjCar);

    ~Map()
    {
        for(int i=0; i < my; i++)
        {
            delete[] MapArray[i];
        }
        delete[] MapArray;
    }
};


void Car::MoveCar(Map *map)
{
    float d_dir = 0.0;
    if(GetAsyncKeyState(VK_UP))
        speed = 0.01;
    else if(GetAsyncKeyState(VK_DOWN))
        speed = -0.01;
    else
        speed = 0.0;

    if(GetAsyncKeyState(VK_LEFT) && speed)
        d_dir = 1.0;
    if(GetAsyncKeyState(VK_RIGHT) && speed)
        d_dir = -1.0;
    dir += d_dir;
    x += cos(dir*PI/180.0)*speed;
    y += sin(dir*PI/180.0)*speed;
    if(map->CollisionCheck(this))
    {
        x -= cos(dir*PI/180.0)*speed;
        y -= sin(dir*PI/180.0)*speed;
        dir -= d_dir;
        speed = 0.0;
    }
}

void Car::DrawCar()
{
    glRotatef(dir, 0.0f, 0.0f, 1.0f);
    glBegin(GL_QUADS);
    glColor3f(1.0f, 1.0f, 0.0f);
    glVertex2f(-lth_half, wth_half);
    glColor3f(0.0f, 0.0f, 1.0f);
    glVertex2f(lth_half, wth_half);
    glColor3f(0.0f, 0.0f, 1.0f);
    glVertex2f(lth_half, -wth_half);
    glColor3f(1.0f, 1.0f, 0.0f);
    glVertex2f(-lth_half, -wth_half);
    glEnd ();
}

float Car::getX()
{
    return x;
}

float Car::getY()
{
    return y;
}
float Car::getDIR()
{
    return dir;
}
float Car::getDIST()
{
    return dist;
}
float Car::getANGLE()
{
    return angle;
}
float Car::getWTH()
{
    return wth_half;
}
float Car::getLTH()
{
    return lth_half;
}


int House::CollissionDetection(Car *ObjCar)
{
    float buff[2];
    for(int i = 0; i < 4; i++)
    {
        ProjectionTo(buff, points[i][0] - ObjCar->getX(), points[i][1] - ObjCar->getY(), ObjCar->getDIR());
        if((fabs(buff[0]) <= wth_half) && (fabs(buff[1]) < lth_half))
            return 1;
    }
    float XDif = ObjCar->getX() - x;
    float YDif = ObjCar->getY() - y;
    float angle;
    for(int i = 0; i < 4; i++)
    {
        buff[0] = XDif;
        buff[1] = YDif;
        if(i < 2)
            angle = ObjCar->getDIR()*PI/180.0+ObjCar->getANGLE();
        else
            angle = ObjCar->getDIR()*PI/180.0-ObjCar->getANGLE();
        if(i%2)
            angle += PI;
        buff[0] += cos(angle)*ObjCar->getDIST();
        buff[1] += sin(angle)*ObjCar->getDIST();
        if((fabs(buff[0]) < wth_half) && (fabs(buff[1]) < lth_half))
            return 1;
    }
    return 0;
}

void House::DrawHouse(Car *ObjCar)
{
    glRotatef(0.0f, 0.0f, 0.0f, 1.0f);
    glBegin(GL_QUADS);
    glColor3f(1.0f, 1.0f, 1.0f);
    glVertex2f(x - ObjCar->getX() - lth_half, y - ObjCar->getY() + wth_half);
    glColor3f(1.0f, 1.0f, 1.0f);
    glVertex2f(x - ObjCar->getX() + lth_half, y - ObjCar->getY() + wth_half);
    glColor3f(1.0f, 1.0f, 1.0f);
    glVertex2f(x - ObjCar->getX() + lth_half, y - ObjCar->getY() - wth_half);
    glColor3f(1.0f, 1.0f, 1.0f);
    glVertex2f(x - ObjCar->getX() - lth_half, y - ObjCar->getY() - wth_half);
    glEnd ();
}




void MapPiece::push(House *NewHouse)
{
    blds[i] = NewHouse;
    i++;
}

int MapPiece::CollisionCheck(Car *ObjCar)
{
    for(int j = 0; j < i; j++)
    {
        if(blds[j]->CollissionDetection(ObjCar))
            return 1;
    }
    return 0;
}

void MapPiece::DrawPiece(Car *ObjCar)
{
    for(int j = 0; j < i; j++)
    {
        blds[j]->DrawHouse(ObjCar);
    }
}




int Map::CollisionCheck(Car *ObjCar)
{
    int x = (int)(ObjCar->getX()/1.0);
    int y = (int)(ObjCar->getY()/1.0);
    for(int i = -1; i < 2; i++)
        for(int j = -1; j < 2; j++)
        {
            if(MapArray[i][j].CollisionCheck(ObjCar))
                return 1;
        }
}

void Map::DrawMap(Car *ObjCar)
{
    int x = (int)(ObjCar->getX()/1.0);
    int y = (int)(ObjCar->getY()/1.0);
    for(int i = -1; i < 2; i++)
        for(int j = -1; j < 2; j++)
        {
            MapArray[i][j].DrawPiece(ObjCar);
        }
}










void DrawCarCoords(Car *obj_car)
{
    glRotatef(0.0f, 0.0f, 0.0f, 0.0f);
    glBegin(GL_LINES);
    glColor3f(0.0f, 1.0f, 0.0f);
    glVertex2f(-obj_car->getX() + 3.1f, -obj_car->getY() + 3.0);
    glColor3f(0.0f, 1.0f, 0.0f);
    glVertex2f(-obj_car->getX() + 2.9f, -obj_car->getY() + 3.0);
    glColor3f(0.0f, 1.0f, 0.0f);
    glVertex2f(-obj_car->getX() + 3.0, -obj_car->getY() + 3.1f);
    glColor3f(0.0f, 1.0f, 0.0f);
    glVertex2f(-obj_car->getX() + 3.0, -obj_car->getY() + 2.9f);
    glEnd();
}




LRESULT CALLBACK WndProc (HWND hWnd, UINT message,
                          WPARAM wParam, LPARAM lParam);
void EnableOpenGL (HWND hWnd, HDC *hDC, HGLRC *hRC);
void DisableOpenGL (HWND hWnd, HDC hDC, HGLRC hRC);

/**************************
 * WinMain
 *
 **************************/

int WINAPI WinMain (HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpCmdLine,
                    int iCmdShow)
{
    WNDCLASS wc;
    HWND hWnd;
    HDC hDC;
    HGLRC hRC;
    MSG msg;
    BOOL bQuit = FALSE;
    float theta = 0.0f;

    /* register window class */
    wc.style = CS_OWNDC;
    wc.lpfnWndProc = WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon (NULL, IDI_APPLICATION);
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH) GetStockObject (BLACK_BRUSH);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = "GLSample";
    RegisterClass (&wc);

    /* create main window */
    hWnd = CreateWindow (
               "GLSample", "OpenGL Sample",
               WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE,
               0, 0, 640, 640,
               NULL, NULL, hInstance, NULL);

    /* enable OpenGL for the window */
    EnableOpenGL (hWnd, &hDC, &hRC);

    /* program main loop */

    Car *OurCar;
    Map MainMap("test map.map", OurCar);
    while (!bQuit)
    {
        /* check for messages */
        if (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
        {
            /* handle or dispatch messages */
            if (msg.message == WM_QUIT)
            {
                bQuit = TRUE;
            }
            else
            {
                TranslateMessage (&msg);
                DispatchMessage (&msg);
            }
        }
        else
        {
            OurCar->MoveCar(&MainMap);
            /* OpenGL animation code goes here */

            glClearColor (1.0f, 0.0f, 0.0f, 0.0f);
            glClear (GL_COLOR_BUFFER_BIT);

            glPushMatrix ();

            MainMap.DrawMap(OurCar);
            OurCar->DrawCar();
            glPopMatrix ();

            SwapBuffers (hDC);
            Sleep (1);
        }
    }

    /* shutdown OpenGL */
    DisableOpenGL (hWnd, hDC, hRC);

    /* destroy the window explicitly */
    DestroyWindow (hWnd);

    return msg.wParam;
}


/********************
 * Window Procedure
 *
 ********************/

LRESULT CALLBACK WndProc (HWND hWnd, UINT message,
                          WPARAM wParam, LPARAM lParam)
{

    switch (message)
    {
    case WM_CREATE:
        return 0;
    case WM_CLOSE:
        PostQuitMessage (0);
        return 0;

    case WM_DESTROY:
        return 0;

    case WM_KEYDOWN:
        switch (wParam)
        {
        case VK_ESCAPE:
            PostQuitMessage(0);
            return 0;
        default:
            return 0;
        }
    default:
        return DefWindowProc (hWnd, message, wParam, lParam);
    }
}


/*******************
 * Enable OpenGL
 *
 *******************/

void EnableOpenGL (HWND hWnd, HDC *hDC, HGLRC *hRC)
{
    PIXELFORMATDESCRIPTOR pfd;
    int iFormat;

    /* get the device context (DC) */
    *hDC = GetDC (hWnd);

    /* set the pixel format for the DC */
    ZeroMemory (&pfd, sizeof (pfd));
    pfd.nSize = sizeof (pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW |
                  PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;
    iFormat = ChoosePixelFormat (*hDC, &pfd);
    SetPixelFormat (*hDC, iFormat, &pfd);

    /* create and enable the render context (RC) */
    *hRC = wglCreateContext( *hDC );
    wglMakeCurrent( *hDC, *hRC );

}


/******************
 * Disable OpenGL
 *
 ******************/

void DisableOpenGL (HWND hWnd, HDC hDC, HGLRC hRC)
{
    wglMakeCurrent (NULL, NULL);
    wglDeleteContext (hRC);
    ReleaseDC (hWnd, hDC);
}
